# THIS IS IN PROGRESS #

from .utils import CplDefObject as obj, STRING_TYPE, INT_TYPE, CplDefArray as arr, FLOAT_TYPE

STRING_ARRAY_TYPE = arr(STRING_TYPE)
INT_ARRAY_TYPE = arr(INT_TYPE)
FLOAT_ARRAY_TYPE = arr(FLOAT_TYPE)
UUID_TYPE = INT_ARRAY_TYPE

ENTITIES_OBJ = obj({})
BLOCKS_Properties = obj({})  # TODO: this

ITEMS__Tag = obj({"Damage": INT_TYPE})  # Not finished
ITEMS__Attribute = obj(
    {"AttributeName": STRING_TYPE, "Amount": FLOAT_TYPE, "Slot": STRING_TYPE, "Operation": INT_TYPE,
     "UUID": UUID_TYPE})
ITEMS__FireworkExplosion = obj(
    {"Colors": INT_ARRAY_TYPE, "FadeColors": INT_ARRAY_TYPE, "Flicker": INT_TYPE, "Trail": INT_TYPE, "Type": INT_TYPE})
ENTITIES__ActiveEffect = obj(
    {"id": STRING_TYPE, "duration": INT_TYPE, "amplifier": INT_TYPE, "ambient": INT_TYPE, "show_particles": INT_TYPE,
     "show_icon": INT_TYPE})
ENTITIES__AttributeModifier = obj({"Name": STRING_TYPE, "Amount": FLOAT_TYPE, "Operation": INT_TYPE, "UUID": UUID_TYPE})
ENTITIES__Attribute = obj({"name": STRING_TYPE, "Base": FLOAT_TYPE, "Modifiers": arr(ENTITIES__AttributeModifier)})

ITEMS_Item = obj({"id": STRING_TYPE, "Count": INT_TYPE, "tag": ITEMS__Tag})
ITEMS_Enchantments = arr(obj({"id": STRING_TYPE, "lvl": INT_TYPE}))
ITEMS_StoredEnchantments = ITEMS_Enchantments
ITEMS_EntityTag = ENTITIES_OBJ
ITEMS_display = obj({"Name": STRING_TYPE, "Lore": STRING_ARRAY_TYPE, "color": INT_TYPE})
ITEMS_AttributeModifiers = arr(ITEMS__Attribute)
ITEMS_Unbreakable = INT_TYPE
ITEMS_SkullOwner = STRING_TYPE
ITEMS_HideFlags = INT_TYPE
ITEMS_CanDestroy = STRING_ARRAY_TYPE
ITEMS_PickupDelay = INT_TYPE
ITEMS_Age = INT_TYPE
ITEMS_generation = INT_TYPE
ITEMS_title = STRING_TYPE
ITEMS_author = STRING_TYPE
ITEMS_pages = STRING_TYPE
ITEMS_Fireworks = obj({"Explosions": arr(ITEMS__FireworkExplosion), "Flight": INT_TYPE})

# Only for blocks as items
ITEMS_CanPlaceOn = STRING_ARRAY_TYPE
ITEMS_BlockEntityTag = obj({})  # Not finished
ITEMS_BlockStateTag = obj({"facing": STRING_TYPE, "half": STRING_TYPE, "shape": STRING_TYPE})

# Entities
ENTITIES_UUID = UUID_TYPE
ENTITIES_Pos = INT_ARRAY_TYPE
ENTITIES_Inventory = arr(ITEMS_Item)
ENTITIES_SelectedItem = ITEMS_Item
ENTITIES_TileEntityData = ENTITIES_OBJ
ENTITIES_Motion = FLOAT_ARRAY_TYPE
ENTITIES_direction = FLOAT_ARRAY_TYPE
ENTITIES_power = FLOAT_ARRAY_TYPE
ENTITIES_ActiveEffects = arr(ENTITIES__ActiveEffect)
ENTITIES_rewardExp = INT_TYPE
ENTITIES_Passengers = arr(ENTITIES_OBJ)
ENTITIES_ArmorItems = arr(ITEMS_Item)
ENTITIES_HandItems = arr(ITEMS_Item)
ENTITIES_HandDropChances = FLOAT_ARRAY_TYPE
ENTITIES_ArmorDropChances = FLOAT_ARRAY_TYPE
ENTITIES_NoAI = INT_TYPE
ENTITIES_NoGravity = INT_TYPE
ENTITIES_Silent = INT_TYPE
ENTITIES_Fire = INT_TYPE
ENTITIES_Invulnerable = INT_TYPE
ENTITIES_Attributes = arr(ENTITIES__Attribute)
ENTITIES_Health = FLOAT_TYPE
ENTITIES_AngerTime = INT_TYPE
ENTITIES_AngryAt = UUID_TYPE
ENTITIES_CustomName = STRING_TYPE
ENTITIES_CustomNameVisible = INT_TYPE
ENTITIES_PersistenceRequired = INT_TYPE
ENTITIES_Type = INT_TYPE
ENTITIES_Saddle = INT_TYPE
ENTITIES_Tame = INT_TYPE
ENTITIES_Variant = INT_TYPE
ENTITIES_Size = INT_TYPE
ENTITIES_BlockState = obj({"Name": STRING_TYPE, "Properties": BLOCKS_Properties})
ENTITIES_Time = INT_TYPE
ENTITIES_DropItem = INT_TYPE
ENTITIES_id = STRING_TYPE
ENTITIES_fuse = INT_TYPE
ENTITIES_ExplosionPower = INT_TYPE
ENTITIES_ExplosionRadius = INT_TYPE
ENTITIES_powered = INT_TYPE
ENTITIES_AttachFace = INT_TYPE
ENTITIES_Peek = INT_TYPE
ENTITIES_APX = INT_TYPE
ENTITIES_APY = INT_TYPE
ENTITIES_APZ = INT_TYPE

# Villager
ENTITIES_Profession = INT_TYPE
ENTITIES_Offers = obj({"Recipes": arr(obj({"buy": ITEMS_Item, "maxUses": INT_TYPE, "sell": ITEMS_Item}))})

# Item Frame
ENTITIES_Facing = INT_TYPE
ENTITIES_ItemRotation = INT_TYPE
ENTITIES_Item = ITEMS_Item
ENTITIES_Invisible = INT_TYPE
ENTITIES_Fixed = INT_TYPE

# Potion
ENTITIES_Potion = STRING_TYPE
ENTITIES__CustomPotionEffect = obj({"Id": STRING_TYPE, "Amplifier": INT_TYPE, "Duration": INT_TYPE})
ENTITIES_CustomPotionEffects = ENTITIES_ActiveEffects
ENTITIES_CustomPotionColor = INT_TYPE

# Armor Stand
# ENTITIES_NoGravity: already defined
ENTITIES_ShowArms = INT_TYPE
ENTITIES_NoBasePlate = INT_TYPE
ENTITIES_Small = INT_TYPE
ENTITIES_Rotation = FLOAT_ARRAY_TYPE
ENTITIES_Marker = INT_TYPE
ENTITIES_Pose = obj(
    {"Head": FLOAT_ARRAY_TYPE, "Body": FLOAT_ARRAY_TYPE, "LeftArm": FLOAT_ARRAY_TYPE, "RightArm": FLOAT_ARRAY_TYPE,
     "LeftLeg": FLOAT_ARRAY_TYPE, "RightLeg": FLOAT_ARRAY_TYPE})
# ENTITIES_Invisible: already defined

# Turtle
ENTITIES_HomePosX = FLOAT_TYPE
ENTITIES_HomePosY = FLOAT_TYPE
ENTITIES_HomePosZ = FLOAT_TYPE
ENTITIES_TravelPosX = FLOAT_TYPE
ENTITIES_TravelPosY = FLOAT_TYPE
ENTITIES_TravelPosZ = FLOAT_TYPE
ENTITIES_HasEgg = INT_TYPE

# Assigning everything to one object type
ENTITIES_OBJ.content = {
    "Pos": ENTITIES_Pos,
    "UUID": ENTITIES_UUID,
    "Inventory": ENTITIES_Inventory,
    "SelectedItem": ENTITIES_SelectedItem,
    "TileEntityData": ENTITIES_TileEntityData,
    "Motion": ENTITIES_Motion,
    "direction": ENTITIES_direction,
    "power": ENTITIES_power,
    "ActiveEffects": ENTITIES_ActiveEffects,
    "rewardExp": ENTITIES_rewardExp,
    "Passengers": ENTITIES_Passengers,
    "ArmorItems": ENTITIES_ArmorItems,
    "HandItems": ENTITIES_HandItems,
    "HandDropChances": ENTITIES_HandDropChances,
    "ArmorDropChances": ENTITIES_ArmorDropChances,
    "NoAI": ENTITIES_NoAI,
    "NoGravity": ENTITIES_NoGravity,
    "Silent": ENTITIES_Silent,
    "Fire": ENTITIES_Fire,
    "Invulnerable": ENTITIES_Invulnerable,
    "Attributes": ENTITIES_Attributes,
    "Health": ENTITIES_Health,
    "AngerTime": ENTITIES_AngerTime,
    "AngryAt": ENTITIES_AngryAt,
    "CustomName": ENTITIES_CustomName,
    "CustomNameVisible": ENTITIES_CustomNameVisible,
    "PersistenceRequired": ENTITIES_PersistenceRequired,
    "Type": ENTITIES_Type,
    "Saddle": ENTITIES_Saddle,
    "Tame": ENTITIES_Tame,
    "Variant": ENTITIES_Variant,
    "Size": ENTITIES_Size,
    "BlockState": ENTITIES_BlockState,
    "Time": ENTITIES_Time,
    "DropItem": ENTITIES_DropItem,
    "id": ENTITIES_id,
    "fuse": ENTITIES_fuse,
    "ExplosionPower": ENTITIES_ExplosionPower,
    "ExplosionRadius": ENTITIES_ExplosionRadius,
    "powered": ENTITIES_powered,
    "AttachFace": ENTITIES_AttachFace,
    "Peek": ENTITIES_Peek,
    "APX": ENTITIES_APX,
    "APY": ENTITIES_APY,
    "APZ": ENTITIES_APZ,
    "Profession": ENTITIES_Profession,
    "Offers": ENTITIES_Offers,
    "Facing": ENTITIES_Facing,
    "ItemRotation": ENTITIES_ItemRotation,
    "Item": ENTITIES_Item,
    "Invisible": ENTITIES_Invisible,
    "Fixed": ENTITIES_Fixed,
    "Potion": ENTITIES_Potion,
    "CustomPotionEffects": ENTITIES_CustomPotionEffects,
    "CustomPotionColor": ENTITIES_CustomPotionColor,
    "ShowArms": ENTITIES_ShowArms,
    "NoBasePlate": ENTITIES_NoBasePlate,
    "Small": ENTITIES_Small,
    "Rotation": ENTITIES_Rotation,
    "Marker": ENTITIES_Marker,
    "Pose": ENTITIES_Pose,
    "HomePosX": ENTITIES_HomePosX,
    "HomePosY": ENTITIES_HomePosY,
    "HomePosZ": ENTITIES_HomePosZ,
    "TravelPosX": ENTITIES_TravelPosX,
    "TravelPosY": ENTITIES_TravelPosY,
    "TravelPosZ": ENTITIES_TravelPosZ,
    "HasEgg": ENTITIES_HasEgg
}

# Command Block
BLOCKS_Command = STRING_TYPE
BLOCKS_auto = INT_TYPE  # Needs redstone toggle

# Generic - Most of the tiles blocks
BLOCKS_CustomName = STRING_TYPE
BLOCKS_Lock = STRING_TYPE

# Beacon
BLOCKS_Primary = STRING_TYPE
BLOCKS_Secondary = STRING_TYPE
BLOCKS_Levels = INT_TYPE

# Entity Spawner
BLOCKS_EntityId = STRING_TYPE
BLOCKS_SpawnData = ENTITIES_OBJ
BLOCKS_SpawnCount = INT_TYPE
BLOCKS_SpawnRange = INT_TYPE
BLOCKS_RequiredPlayerRange = INT_TYPE
BLOCKS_Delay = INT_TYPE
BLOCKS_MinSpawnDelay = INT_TYPE
BLOCKS_MaxSpawnDelay = INT_TYPE
BLOCKS_MaxNearbyEntities = INT_TYPE
BLOCKS_SpawnPotentials = arr(obj({"data": ENTITIES_OBJ, "weight": INT_TYPE}))
BLOCKS_Weight = INT_TYPE



from enum import Enum
from typing import Dict, Union, List, Any

FLOAT_PREC = 1000
INT_LIMIT = 2147483647


def get_float_limit():
    return INT_LIMIT / FLOAT_PREC


_expr_id = 0

VERSION_RADON = "1.9.1"


def basic_calc(a: Union[int, float], op: str, b: Union[int, float]) -> int | float:
    if op == "+":
        return a + b
    elif op == "-":
        return a - b
    elif op == "*":
        return a * b
    elif op == "/":
        return a / b
    elif op == "%":
        return a % b
    else:
        assert False


def basic_cmp(a, op: str, b):
    if op == ">":
        return a > b
    elif op == "<":
        return a < b
    elif op == ">=":
        return a >= b
    elif op == "<=":
        return a <= b
    else:
        assert False


def inv_cmp(op: str):
    if op == ">":
        return "<="
    elif op == "<":
        return ">="
    elif op == ">=":
        return "<"
    elif op == "<=":
        return ">"
    elif op == "==":
        return "!="
    elif op == "!=":
        return "=="
    else:
        assert False


# Version argument can be: a pack format, a minecraft version like 1.16.5, or a snapshot version like 23w32a
def get_pack_format(version):
    if not version:
        return None
    if isinstance(version, int):
        return version
    if version.isnumeric() and (4 <= int(version) <= 18 or version == "48"):
        return int(version)
    version = version.split("-")[0]
    if len(version) > 2 and version[2] == "w":
        version = version[:-1]
        spl = version.split("w")
        v = spl[0]
        if not v.isnumeric():
            return None
        v = int(v)
        k = spl[1]
        if not k.isnumeric():
            return None
        k = int(k)
        if v < 17 or (v == 17 and k < 48):
            return None
        if v < 20:
            return 4
        if v < 21:
            return 7
        if v < 22:
            return 8
        if v < 23:
            return 10
        if v < 24:
            if k < 6:
                return 11
            if k < 12:
                return 12
            if k < 16:
                return 13
            if k < 18:
                return 14
            if k < 31:
                return 15
            if k < 32:
                return 16
            if k > 35:
                return None
            return 17
        return None
    spl = version.split(".")
    if spl[0] != "1" or len(spl) == 1 or len(spl) > 3:
        return None
    v = spl[1]
    if not v.isnumeric():
        return None
    v = int(v)
    k = spl[2] if len(spl) == 3 else "0"
    if not k.isnumeric():
        return None
    k = int(k)
    if v > 21 or k > 100:
        return None
    vk = v * 100 + k
    if vk < 1300:
        return None
    if vk < 1500:
        return 4
    if vk < 1602:
        return 5
    if vk < 1700:
        return 6
    if vk < 1800:
        return 7
    if vk < 1802:
        return 8
    if vk < 1900:
        return 9
    if vk < 1904:
        return 10
    if vk < 2000:
        return 12
    if vk < 2002:
        return 15
    if vk < 2100:
        return 18
    return 48


def reset_expr_id():
    global _expr_id
    _expr_id = 0


def get_uuid():
    global _expr_id
    _expr_id += 1
    return _expr_id


class TokenType(Enum):
    KEYWORD = "keyword"
    IDENTIFIER = "identifier"
    STRING_LITERAL = "string_literal"
    INT_LITERAL = "int_literal"
    FLOAT_LITERAL = "float_literal"
    OPERATOR = "operator"
    WHITESPACE = "whitespace"
    SYMBOL = "symbol"
    EOL = "eol"  # end of line, new line
    EOF = "eof"  # end of file
    EOE = "eoe"  # end of expression, semicolon

    GROUP = "group"
    FUNCTION_CALL = "function_call"
    LAMBDA_FUNCTION = "lambda_function"
    SELECTOR = "selector"
    SELECTOR_IDENTIFIER = "selector_identifier"
    BLOCK_IDENTIFIER = "block_identifier"

    POINTER = "pointer"  # This is not a token type. This is used to point to a part of the user code.


class UniversalStrMixin:
    def __str__(self):
        attributes = [
            f"{key}={list(map(str, value)) if isinstance(value, list) else value}"
            for key, value in filter(lambda x: x[0] != "code", self.__dict__.items())
        ]
        return f"{self.__class__.__name__}({', '.join(attributes)})"


class CplDef:
    def __init__(self, type: str):
        self.type = type

    def __eq__(self, other) -> bool:
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    def get_sample_value(self) -> str:
        return ""


class VariableDeclaration:
    def __init__(self, name: str, type: Any, constant: bool, init_func=None, get_func=None):
        self.name = name
        self.value = None
        if not type.__class__.__name__.startswith("CplDef"):
            self.value = type
            type = type.unique_type
        self.type = type
        self.constant = constant
        self.init_func = init_func
        self.get_func = get_func


class CplDefInt(CplDef):
    def __init__(self):
        super().__init__("int")

    def __str__(self) -> str:
        return "int"

    def get_sample_value(self) -> str:
        return "0"


class CplDefFloat(CplDef):
    def __init__(self):
        super().__init__("float")

    def __str__(self) -> str:
        return "float"

    def get_sample_value(self) -> str:
        return "0.0"


class CplDefString(CplDef):
    def __init__(self):
        super().__init__("string")

    def __str__(self) -> str:
        return "string"

    def get_sample_value(self) -> str:
        return "''"


class CplDefArray(CplDef):
    def __init__(self, content: CplDef):
        super().__init__("array")
        self.content = content

    def __str__(self) -> str:
        return str(self.content) + "[]"

    def get_sample_value(self) -> str:
        return f"[{self.content.get_sample_value()}]"


class CplDefTuple(CplDef):
    def __init__(self, content: List[CplDef]):
        super().__init__("tuple")
        self.content = content

    def __str__(self) -> str:
        return "[" + ", ".join(map(str, self.content)) + "]"

    def get_sample_value(self) -> str:
        return f"[{', '.join(map(lambda x: x.get_sample_value(), self.content))}]"


class CplDefObject(CplDef):
    def __init__(self, type: Dict[str, CplDef], class_name: str | None = None):
        super().__init__("object")
        self.content = type
        self.class_name = class_name

    def __str__(self) -> str:
        if self.class_name:
            return self.class_name
        return (
                "{"
                + ", ".join([str(k) + ": " + str(v) for k, v in self.content.items()])
                + "}"
        )

    def get_sample_value(self) -> str:
        return (
                "{"
                + ", ".join([k + ": " + self.content[k].get_sample_value() for k in self.content])
                + "}"
        )


class CplDefSelector(CplDef):
    def __init__(self):
        super().__init__("selector")

    def __str__(self) -> str:
        return "selector"

    def get_sample_value(self) -> str:
        raise "Selectors cannot be sampled"


class CplDefFunction(CplDef):
    def __init__(self, arguments: List[CplDef], returns: CplDef | None):
        super().__init__("string")
        self.arguments = arguments
        self.returns = returns

    def __str__(self) -> str:
        return "((" + ", ".join(map(str, self.arguments)) + ") => " + str(self.returns or "void") + ")"

    def get_sample_value(self) -> str:
        raise "Functions cannot be sampled"


INT_TYPE = CplDefInt()
FLOAT_TYPE = CplDefFloat()
STRING_TYPE = CplDefString()
SELECTOR_TYPE = CplDefSelector()

str_counter = {}


def get_str_count(s):
    if s in str_counter:
        return str_counter[s]
    str_counter[s] = get_uuid()
    return str_counter[s]



class RadonError(Exception):
    def __init__(self, err, text, code, start, end):
        super().__init__(err)
        self.err = err
        self.text = text
        self.code = code
        self.start = start
        self.end = end


def show_err(color, text, code, start, end):
    lines = code.split("\n")

    line_start = code[:start].count("\n")
    line_end = code[:end].count("\n")
    col_start = start - (code.rfind("\n", 0, start) + 1)
    col_end = end - (code.rfind("\n", 0, end) + 1)

    if line_start == line_end:
        ls = lines[line_start]
        lines[line_start] = (
                ls[:col_start]
                + "\033[" + color + "m\033[4m"
                + ls[col_start:col_end]
                + "\033[0m" + "\033[" + color + "m"
                + ls[col_end:]
        )
    else:
        ls = lines[line_start]
        lines[line_start] = (
                ls[:col_start]
                + "\033[" + color + "m\033[4m"
                + ls[col_start:]
                + "\033[0m"
        )
        for i in range(line_start + 1, line_end):
            lines[i] = "\033[" + color + "m\033[4m" + lines[i] + "\033[0m"
        lines[line_end] = (
                "\033[" + color + "m\033[4m"
                + lines[line_end][:col_end]
                + "\033[0m"
                + lines[line_end][col_end:]
        )

    line_inc = 0 if color == "33" else 2
    start_line_index = max(0, line_start - line_inc)
    end_line_index = min(len(lines) - 1, line_end + line_inc)

    printing = ""

    for i in range(start_line_index, end_line_index + 1):
        prefix = f"{i + 1} | "
        printing += prefix + lines[i] + ("\n" if color == "31" else "")

    printing += text

    return printing


def raise_syntax_error(text, token):
    raise_error("SyntaxError", text, token)


def raise_syntax_error_t(text, code, start, end):
    raise_error_t("SyntaxError", text, code, start, end)


def raise_error(error, text, token):
    # print("Token: " + str(token))
    token.end += 1
    raise_error_t(error, text, token.code, token.start, token.end)


def raise_error_t(error, text, code, start, end):
    raise RadonError(
        show_err("31", f"\n\033[31m" + (f"{error}: " if error else "") + f"{text}\033[0m", code, start, end),
        text, code, start, end)


def show_warning(text, token):
    # print("Token: " + str(token))
    print(
        show_err("33",
                 f"\n\033[33mWarning: {text}\033[0m", token.code, token.start, token.end
                 )
    )



from typing import List

from ..error import raise_error
from ..tokenizer import GroupToken
from ..transpiler import FunctionDeclaration, TranspilerContext, add_lib, py_to_cpl


def lib_eval(_ctx: TranspilerContext, _args: List[GroupToken], token: GroupToken):
    try:
        return py_to_cpl(eval(token.value[7:-1]))
    except Exception as e:
        raise_error("Eval error", str(e), token)
        return None


def lib_exec(_ctx: TranspilerContext, _args: List[GroupToken], token: GroupToken):
    try:
        _globals = {"result": 0}
        exec(token.value[7:-1], _globals)
        return py_to_cpl(_globals["result"])
    except Exception as e:
        raise_error("Exec error", str(e), token)


add_lib(FunctionDeclaration(
    type="python-raw",
    name="pyeval",
    returns="void",
    function=lib_eval,
))

add_lib(FunctionDeclaration(
    type="python-raw",
    name="pyexec",
    returns="void",
    function=lib_exec,
))



import json
import re

from ..cpl.int import CplInt
from ..cpl.string import CplString
from ..error import raise_syntax_error
from ..transpiler import add_lib, CustomCplObject
from ..utils import VariableDeclaration, CplDefFunction, STRING_TYPE, get_str_count, get_uuid

funcArgs = [CplDefFunction([], None)]  # (() => void)


def first_join_init(tr, file_name):
    tr.dp_files["advancements/__events__/on_first_join.json"] = json.dumps({
        "criteria": {
            "requirement": {
                "trigger": "minecraft:tick"
            }
        },
        "rewards": {
            "function": f"{tr.pack_namespace}:{file_name}"
        }
    }, indent=2)


def join_init(tr, file_name):
    tr.files[file_name].insert(0, f"scoreboard players operation @s on_join = counter? on_join")
    tr.load_file.append(f"scoreboard objectives add on_join dummy")
    tr.tick_file = [
                       "scoreboard players add counter? on_join 1",
                       "scoreboard players add @a on_join 1",
                       f"execute as @a unless score @s on_join = counter? on_join run function {tr.pack_namespace}:{file_name}"
                   ] + tr.tick_file


def die_common(tr):
    tr.load_file.append(f"scoreboard objectives add on_die deathCount")
    tr.tick_file = [
                       "execute as @a[scores={on_die=1}] at @s run function " + f"{tr.pack_namespace}:__events__/die",
                       "execute as @e[type=player,scores={__die__=2..}] at @s run function namespace:__events__/respawn"
                   ] + tr.tick_file


def die_init(tr, file_name):
    tr.files[file_name].insert(0, f"scoreboard players set @s on_die 2")
    respawn_file_name = f"__events__/respawn"
    if respawn_file_name not in tr.files:
        tr.files[respawn_file_name] = [f"scoreboard players set @s on_die 0"]
    die_common(tr)


def respawn_init(tr, file_name):
    tr.files[file_name].insert(0, f"scoreboard players set @s on_die 0")
    die_file_name = f"__events__/die"
    if die_file_name not in tr.files:
        tr.files[die_file_name] = [f"scoreboard players set @s on_die 2"]
    die_common(tr)


def stick_init(tr, file_name, item):
    tr.load_file.append(f"scoreboard objectives add on_{item} used:{item}")
    tr.tick_file.insert(0,
                        "execute as @a[scores={on_" + item + "=1..}] at @s run function " + f"{tr.pack_namespace}:{file_name}")

    tr.files[file_name].insert(0, f"scoreboard players set @s on_{item} 0")


def listener_on(ctx, arguments, token):
    tr = ctx.transpiler
    if ctx.function or ctx.loop:
        raise_syntax_error("Listener.on() can only be called in the main scope", token)
    tr.assert_args("Listener.on", [
        STRING_TYPE, CplDefFunction([], None)
    ], arguments, token)
    arg0 = arguments[0]
    arg1 = arguments[1]
    if not isinstance(arg0, CplString) or not isinstance(arg1, CplString):
        raise_syntax_error("Expected a literal string as the first argument for Player.on()", token)
    event = arg0.value
    event = event.lower()
    obj_type = event

    if event == "rejoin":
        obj_type = "custom:leave_game"
    if event == "sneak":
        obj_type = "custom:sneak_time"
    if event == "sprint":
        obj_type = "custom:sprint_one_cm"
    # only a-z allowed, use regex
    event_str = str(get_str_count(event))
    file_name = (f"__events__/{event}"
                 if re.fullmatch(r"[a-z_]+", event)
                 else f"__events__/{event_str}")

    if file_name not in tr.files:

        tr.files[file_name] = []
        if event == "firstjoin":
            first_join_init(tr, file_name)
        elif event == "join":
            join_init(tr, file_name)
        elif event == "die":
            die_init(tr, file_name)
        elif event == "respawn":
            respawn_init(tr, file_name)
        elif event == "carrot_on_a_stick" or event == "warped_fungus_on_a_stick":
            stick_init(tr, file_name, event)
        else:
            tr.files[file_name].insert(0, f"scoreboard players set @s on_{event_str} 0")
            tr.load_file.append(f"scoreboard objectives add on_{event_str} {obj_type}")
            tr.tick_file.insert(0,
                                "execute as @a[scores={on_" + event_str + "=1..}] at @s run function " + tr.pack_namespace + ":" + file_name)

    arg1_fn = tr.get_fn(arg1.value, [], token)

    eid = get_uuid()

    tr.files[file_name].append(
        f"execute if score __event__{event_str}_{eid} __temp__ matches 1..1 run function {tr.pack_namespace}:{arg1_fn.file_name}")

    ctx.file.append(f"scoreboard players set __event__{event_str}_{eid} __temp__ 1")

    return CplInt(token, 0)  # returns 0 as an int


add_lib(VariableDeclaration(
    name="Listener",
    type=CustomCplObject({
        "on": listener_on
    }),
    constant=True
))



import json
from typing import List

from ..cpl._base import CompileTimeValue
from ..cpl.int import CplInt
from ..cpl.object import CplObject
from ..error import raise_syntax_error
from ..transpiler import add_lib, TranspilerContext, CustomCplObject
from ..utils import VariableDeclaration, get_uuid


def lib_recipe_add(ctx: TranspilerContext, arguments: List[CompileTimeValue], token):
    tr = ctx.transpiler
    if len(arguments) < 1:
        raise_syntax_error("Recipe.table() requires at least 1 argument", token)
    recipe = arguments[0]
    if not isinstance(recipe, CplObject):
        raise_syntax_error("Expected an object for the first argument of Recipe.table()", token)
    recipe = recipe.get_py_value()
    if not recipe:
        raise_syntax_error("Expected the table to be a literal object", token)
    tr.dp_files[f"recipe{tr.s}/recipe_{get_uuid()}.json"] = json.dumps(recipe, indent=2)
    return CplInt(token, 0)


add_lib(VariableDeclaration(
    name="Recipe",
    type=CustomCplObject({
        "add": lib_recipe_add
    }),
    constant=True
))



import json
from typing import List

from ._base import CompileTimeValue
from ..error import raise_syntax_error
from ..tokenizer import Token
from ..utils import CplDefArray


class CplArray(CompileTimeValue):
    def __init__(self, token, value):
        # type: (Token | None, List[CompileTimeValue]) -> None
        self.unique_type: CplDefArray
        super().__init__(token)
        self.value = value
        if len(value) == 0:
            raise_syntax_error("Array cannot be empty", token)
        self.unique_type = CplDefArray(value[0].unique_type)
        for v in value:
            if v.unique_type != value[0].unique_type:
                raise_syntax_error("All array elements must be of the same type", token)

    def get_py_value(self):
        ls = []
        for v in self.value:
            vl = v.get_py_value()
            if vl is None:
                return None
            ls.append(vl)
        return ls

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        if force == "score":
            return None
        py_val = self.get_py_value()
        if py_val is not None:
            ctx.file.append(f"data modify {nbt_loc} set value {json.dumps(py_val)}")
        else:
            init = []
            keys = []

            for index, c in enumerate(self.value):
                c_py_val = c.get_py_value()
                if c_py_val is not None:
                    init.append(json.dumps(c_py_val))
                else:
                    init.append(c.unique_type.get_sample_value())
                    keys.append(index)
            ctx.file.append(f"data modify {nbt_loc} set value [{','.join(init)}]")

            for index, v in enumerate(self.value):
                if index in keys:
                    v.cache(ctx, nbt_loc=f"{nbt_loc}[{index}]", force="nbt")
        return CplArrayNBT(self.token, nbt_loc, self.unique_type)

    def _get_index(self, ctx, index: CompileTimeValue):
        if isinstance(index, CplString) and index.value == "length":
            return CplInt(self.token, len(self.value))
        if isinstance(index, CplInt) or isinstance(index, CplString):
            try:
                return self.value[int(index.value)]
            except IndexError:
                return None
        if isinstance(index, CplFloat):
            return None
        return self.cache(ctx).get_index(ctx, index)

    def _add(self, ctx, cpl):
        if isinstance(cpl, CplArray):
            return CplArray(self.token, self.value + cpl.value)
        if not isinstance(cpl, CplArrayNBT):
            return None
        return self.cache(ctx)._add(ctx, cpl)

    def tellraw_object(self, ctx):
        py_v = self.get_py_value()
        if py_v is None:
            return self.cache(ctx).tellraw_object(ctx)
        return {"text": json.dumps(py_v)}


from .float import CplFloat
from .int import CplInt
from .nbtarray import CplArrayNBT
from .string import CplString



import json
from typing import List

from ._base import CompileTimeValue
from .float import CplFloat
from .int import CplInt
from .string import CplString
from ..error import raise_syntax_error
from ..tokenizer import Token
from ..utils import CplDefArray, CplDefTuple


class CplTuple(CompileTimeValue):
    def __init__(self, token, value):
        # type: (Token | None, List[CompileTimeValue]) -> None
        self.unique_type: CplDefArray
        super().__init__(token)
        self.value = value
        self.unique_type = CplDefTuple(list(map(lambda x: x.unique_type, value)))
        ls = []
        for v in value:
            vl = v.get_py_value()
            if vl is None:
                raise_syntax_error("Tuples(constant arrays) cannot have non-literal values", token)
            ls.append(vl)
        self._py_val = ls

    def get_py_value(self):
        return self._py_val

    def _get_index(self, ctx, index: CompileTimeValue):
        if isinstance(index, CplString) and index.value == "length":
            return len(self.value)
        if isinstance(index, CplInt) or isinstance(index, CplString):
            try:
                return self.value[int(index.value)]
            except IndexError:
                return None
        if isinstance(index, CplFloat):
            return None
        return None

    def _get_slice(self, ctx, index1, index2, index3, token):
        # TODO: this
        return None

    def tellraw_object(self, ctx):
        return {"text": json.dumps(self._py_val)}



import json
from typing import Union, List

from ._base import CompileTimeValue
from ._literal import CplLiteral
from ..tokenizer import Token
from ..utils import STRING_TYPE, TokenType, get_uuid


class CplString(CplLiteral):
    def __init__(self, token, value, is_fn_reference=False):
        # type: (Union | None, str, bool) -> None
        value = str(value)
        if token is None:
            token = Token(str(value), TokenType.STRING_LITERAL, 0, len(str(value)))
        super().__init__(token, value)
        self.unique_type = STRING_TYPE
        self.is_fn_reference = is_fn_reference

    def get_py_value(self):
        return self.value

    def _call(self, ctx, arguments):
        if not self.is_fn_reference:
            return None
        return ctx.transpiler.run_function_with_cpl(ctx, self.value, arguments, self.token)

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        if force == "score":
            raise ValueError("Cannot store string as score")
        ctx.file.append(f"data modify {nbt_loc} set {self.get_data_str(ctx)}")
        return CplStringNBT(self.token, nbt_loc)

    def get_data_str(self, ctx):
        return f"value {json.dumps(self.value)}"

    def _get_index(self, ctx, index: CompileTimeValue):
        if isinstance(index, CplString) and index.value == "length":
            return CplInt(self.token, len(self.value))
        return None

    def _get_slice(self, ctx, index1, index2, index3, token):
        if isinstance(index1, CplInt) and isinstance(index2, CplInt) and isinstance(index3, CplInt):
            return CplString(self.token, self.value[index1.value:index2.value:index3.value])
        return self.cache(ctx)._get_slice(ctx, index1, index2, index3, token)

    def _call_index(self, ctx, index: str, arguments: List[CompileTimeValue], token):
        if index == "toString":
            if len(arguments) > 0:
                return 0
            return self
        return None

    def _add(self, ctx, cpl):
        if isinstance(cpl, CplString):
            return CplString(self.token, self.value + cpl.value)
        if not isinstance(cpl, CplStringNBT):
            return None
        eid = f"storage temp _{get_uuid()}"
        file_name = ctx.transpiler.get_temp_file_name(
            f"$data modify {eid} set value '{self.value}$(_0)'")
        cpl.cache(ctx, nbt_loc="string_concat _0", force="nbt")
        ctx.file.append(f"function {ctx.transpiler.pack_namespace}:{file_name} with storage string_concat")
        return CplStringNBT(self.token, eid)

    def _and(self, ctx, cpl):
        return cpl._and(ctx, CplInt(self.token, len(self.value)))

    def _or(self, ctx, cpl):
        return cpl._or(ctx, CplInt(self.token, len(self.value)))

    def tellraw_object(self, ctx):
        return {"text": self.value}


from .int import CplInt
from .nbtstring import CplStringNBT



from enum import Enum
from typing import List, Tuple, Union

from .error import raise_syntax_error
from .tokenizer import (
    GroupToken,
    Token,
    TokenType,
    cpl_def_from_tokens,
    is_float,
    split_tokens,
    tokenize, CMP_OP, CMP_COMBINE_OP,
)
from .utils import UniversalStrMixin

ENDERS = [
    TokenType.EOL,
    TokenType.EOE,
    TokenType.EOF,
]

EXECUTE_MACROS = [
    "align",
    "anchored",
    "as",
    "at",
    "facing",
    "in",
    "on",
    "positioned",
    "rotated",
    "summon",
    "if",
    "unless",
    "store",
]

COMMANDS = [
    "ability",
    "advancement",
    "attribute",
    "ban",
    "ban-ip",
    "banlist",
    "bossbar",
    "clear",
    "clone",
    "damage",
    "data",
    "datapack",
    "debug",
    "defaultgamemode",
    "deop",
    "difficulty",
    "effect",
    "enchant",
    "execute",
    "experience",
    "xp",
    "fill",
    "fillbiome",
    "forceload",
    "function",
    "gamemode",
    "gamerule",
    "give",
    "item",
    "jfr",
    "kick",
    "kill",
    "list",
    "locate",
    "loot",
    "me",
    "msg",
    "tell",
    "w",
    "op",
    "pardon",
    "pardon-ip",
    "particle",
    "perf",
    "permission",
    "ops",
    "place",
    "playsound",
    "publish",
    "random",
    "sequence",
    "recipe",
    "reload",
    "return",
    "ride",
    "save-all",
    "save-off",
    "save-on",
    "say",
    "schedule",
    "scoreboard",
    "seed",
    "setblock",
    "setidletimeout",
    "setworldspawn",
    "spawnpoint",
    "spectate",
    "spreadplayers",
    "stop",
    "stopsound",
    "summon",
    "tag",
    "team",
    "teleport",
    "tp",
    "tell",
    "msg",
    "w",
    "tellraw",
    "time",
    "title",
    "tm",
    "teammsg",
    "tp",
    "teleport",
    "trigger",
    "w",
    "tell",
    "msg",
    "weather",
    "whitelist",
    "worldborder",
    "xp",
    "experience",
]

operators = {
    "**": {
        "p": 4,
        "a": "right",
    },
    "*": {
        "p": 3,
        "a": "left",
    },
    "/": {
        "p": 3,
        "a": "left",
    },
    "%": {
        "p": 3,
        "a": "left",
    },
    "+": {
        "p": 2,
        "a": "left",
    },
    "-": {
        "p": 2,
        "a": "left",
    },
    "not": {
        "p": 2,
        "a": "left",
    },
    "is": {
        "p": 1,
        "a": "left",
    },
    "is not": {
        "p": 1,
        "a": "left",
    },
    "==": {
        "p": 1,
        "a": "left",
    },
    "!=": {
        "p": 1,
        "a": "left",
    },
    ">": {
        "p": 1,
        "a": "left",
    },
    "<": {
        "p": 1,
        "a": "left",
    },
    ">=": {
        "p": 1,
        "a": "left",
    },
    "<=": {
        "p": 1,
        "a": "left",
    },
    "or": {
        "p": 0,
        "a": "left",
    },
    "and": {
        "p": 0,
        "a": "left",
    },
    "||": {
        "p": 0,
        "a": "left",
    },
    "&&": {
        "p": 0,
        "a": "left",
    },
}

EXPR_EXPR = [
    TokenType.FUNCTION_CALL,
    TokenType.GROUP,
    TokenType.INT_LITERAL,
    TokenType.FLOAT_LITERAL,
    TokenType.STRING_LITERAL,
    TokenType.IDENTIFIER,
    TokenType.SELECTOR_IDENTIFIER,
    TokenType.SELECTOR,
    TokenType.LAMBDA_FUNCTION
]

EXPR_OP = list("+-*/%") + ["**"]


def is_time(token: Token):
    return (
            token.type == TokenType.IDENTIFIER
            and token.value[-1] in list("dst")
            and is_float(token.value[:-1])
    )


class StatementType(Enum):
    INLINE = "inline"
    SET_VARIABLE = "set_variable"
    DEFINE_FUNCTION = "define_function"
    RETURN = "return"
    BREAK = "break"
    CONTINUE = "continue"
    INTRODUCE_VARIABLE = "introduce_variable"
    EXECUTE_MACRO = "execute_macro"
    IF_FLOW = "if_flow"
    FOR = "for"
    WHILE = "while"
    LOOP = "loop"
    SCHEDULE = "schedule"
    IMPORT = "import"
    DEFINE_CLASS = "define_class"


class Statement(UniversalStrMixin):
    def __init__(self, type: StatementType, code: str, start: int, end: int):
        self.type = type
        self.code = code
        self.start = start
        self.end = end


class ImportStatement(Statement):
    def __init__(
            self, code: str, start: int, end: int, path: Token, as_: Token | None = None
    ):
        super().__init__(StatementType.IMPORT, code, start, end)
        self.path = path
        self.as_ = as_


class ScheduleStatement(Statement):
    def __init__(
            self, code: str, start: int, end: int, time: Token, body: List[Statement]
    ):
        super().__init__(StatementType.SCHEDULE, code, start, end)
        self.time = time
        self.body = body


class IfFlowStatement(Statement):
    def __init__(
            self,
            code: str,
            start: int,
            end: int,
            condition: List[Token],
            body: List[Statement],
            else_body: List[Statement] | None,
    ):
        super().__init__(StatementType.IF_FLOW, code, start, end)
        self.condition = condition
        self.body = body
        self.elseBody = else_body


class LoopStatement(Statement):
    def __init__(
            self, code: str, start: int, end: int, time: Token | None, body: List[Statement], step: List[Statement]
    ):
        super().__init__(StatementType.LOOP, code, start, end)
        self.time = time
        self.body = body
        self.step = step


class IntroduceVariableStatement(Statement):
    def __init__(
            self, code: str, start: int, end: int, name: Token, var_type: List[Token]
    ):
        super().__init__(StatementType.INTRODUCE_VARIABLE, code, start, end)
        self.name = name
        self.varType = var_type


class DefineFunctionStatement(Statement):
    def __init__(
            self,
            code: str,
            start: int,
            end: int,
            name: Token,
            body: List[Statement],
            arguments: List[List[List[Token]]],
            returns: Union[List[Token], str],
    ):
        super().__init__(StatementType.DEFINE_FUNCTION, code, start, end)
        self.name = name
        self.body = body
        self.arguments = (
            arguments  # [ [ [TypeOfTheArgument...], [ NameOfTheArgument ] ] ]
        )
        self.returns = returns


class ReturnStatement(Statement):
    def __init__(
            self, code: str, start: int, end: int, keyword: Token, expr_tokens: List[Token]
    ):
        super().__init__(StatementType.RETURN, code, start, end)
        self.keyword = keyword
        self.exprTokens = expr_tokens


class BreakStatement(Statement):
    def __init__(self, code: str, start: int, end: int, keyword: Token):
        super().__init__(StatementType.BREAK, code, start, end)
        self.keyword = keyword


class ContinueStatement(Statement):
    def __init__(self, code: str, start: int, end: int, keyword: Token):
        super().__init__(StatementType.CONTINUE, code, start, end)
        self.keyword = keyword


class InlineStatement(Statement):
    def __init__(self, code: str, start: int, end: int, expr_tokens: List[Token]):
        super().__init__(StatementType.INLINE, code, start, end)
        self.exprTokens = expr_tokens


class ExecuteMacroStatement(Statement):
    def __init__(
            self, code: str, start: int, end: int, command: str, body: List[Statement]
    ):
        super().__init__(StatementType.EXECUTE_MACRO, code, start, end)
        self.command = command
        self.body = body


class DefineClassStatement(Statement):
    def __init__(
            self,
            code: str,
            start: int,
            end: int,
            name: Token,
            attributes: List[Tuple[str, List[Token]]] = None,
            methods: List[DefineFunctionStatement] = None,
    ):
        super().__init__(StatementType.DEFINE_CLASS, code, start, end)
        if attributes is None:
            attributes = []
        if methods is None:
            methods = []
        self.name = name
        self.attributes = attributes
        self.methods = methods


def read_until_curly(tokens: List[Token], start_index: int):
    length = len(tokens)
    index = start_index
    while index < length:
        t = tokens[index]
        if isinstance(t, GroupToken) and t.open.value == "{":
            return index
        index += 1
    return None


def read_until_types(tokens: List[Token], start_index: int, types: List[TokenType]):
    length = len(tokens)
    index = start_index
    while index < length:
        t = tokens[index]
        if t.type in types:
            return index
        index += 1
    return None


def read_expression(tokens: List[Token], start_index: int):
    length = len(tokens)
    index = start_index
    result = []
    if len(tokens) - 1 < index:
        return index, result

    while index < length:
        token = tokens[index]

        if token.type in {TokenType.EOF, TokenType.EOE}:
            break
        elif token.type == TokenType.EOL:
            tl = None if index == 0 else tokens[index - 1]
            if not tl or tl.value != "\\":
                break
            result.pop()
            index += 1
            continue

        result.append(token)
        index += 1

    return index, result


def chain_tokens_iterate(tokens: List[Token], index: List[int]) -> List[Token]:
    """
    Groups the first single token chain at the given index.

    Example return:
    Input: (a.d).b[c].test() + 1
    [GroupToken('(a.d)'), IdentifierToken('b'), GroupToken('[c]'), FunctionCallToken('test()')]
    """
    t0 = next_token(tokens, index)
    if t0 is None:
        return []
    branch = [t0]
    if t0.type not in {
        TokenType.GROUP,
        TokenType.IDENTIFIER,
        TokenType.BLOCK_IDENTIFIER,
        TokenType.SELECTOR_IDENTIFIER,
        TokenType.INT_LITERAL,
        TokenType.FLOAT_LITERAL,
        TokenType.STRING_LITERAL,
        TokenType.FUNCTION_CALL,
        TokenType.SELECTOR,
    } or (isinstance(t0, GroupToken) and t0.open.value not in {"(", "["}):
        return branch
    while True:
        tl = tokens[index[0]]
        tn = next_token(tokens, index)
        if tn is None:
            return branch
        t_next = peek_token(tokens, index)
        if tn.type == TokenType.EOL and t_next and t_next.value == ".":
            index[0] += 1
            continue
        if tn.value == ".":
            if tl.value == ".":
                index[0] -= 2
                return branch
            continue
        if (
                (
                        tn.type == TokenType.FUNCTION_CALL
                        or tn.type == TokenType.IDENTIFIER
                        or tn.type == TokenType.BLOCK_IDENTIFIER
                        or tn.type == TokenType.SELECTOR_IDENTIFIER
                        or tn.type == TokenType.INT_LITERAL
                        or tn.type == TokenType.FLOAT_LITERAL
                        or tn.type == TokenType.SELECTOR
                )
                and tl.value == "."
                and tl.end == tn.start
        ):
            branch.append(tn)
            continue
        if isinstance(tn, GroupToken) and tn.open.value in {"[", "("}:
            branch.append(tn)
            continue
        index[0] -= 1
        return branch


def chain_tokens(tokens: List[Token]) -> List[List[Token]]:
    """
    Groups the indexing tokens into a single token.

    Things like:
    (a.b[c].d.f).e
    [1, 2, 3][0]
    a.b.test(5, 6, 7)[0].b

    Example return:
    Input: (a.d).b[c].test() + 1
    Output: [
        [GroupToken('(a.d)'), IdentifierToken('b'), GroupToken('[c]'), FunctionCallToken('test()')],
        [Token('+')],
        [Token('1')]
    ]

    See that GroupToken('a.d') stayed as it is because it will be dealt with in the next step
    """
    if len(tokens) == 0:
        return []
    chains = []
    index = [-1]

    while True:
        if len(tokens) > index[0] + 1 and tokens[index[0] + 1].type == TokenType.EOF:
            return chains
        chain = chain_tokens_iterate(tokens, index)
        if len(chain) == 0:
            return chains
        chains.append(chain)


def make_expr(chains: List[List[Token]]) -> List[List[Token]]:
    """
    Converts the given token chain group into an easily executable expression.
    Example:

    Input: [Token('a'), Token('+'), Token('b')] or a + b
    Output: [Token('a'), Token('b'), Token('+')] or a b +

    For more information: https://en.wikipedia.org/wiki/Shunting_yard_algorithm
    """
    new_chains = []
    last_op = False
    for index, chain in enumerate(chains):
        if chain[0].type in ENDERS:
            continue
        if chain[0].type not in EXPR_EXPR:
            if last_op or index == 0:
                if chain[0].value != "-":
                    raise_syntax_error("Expected a non-operator non-symbol token", chain[0])
                new_chains.insert(-2, [Token("0", TokenType.INT_LITERAL, 0, 1)])
            last_op = True
        last_op = False

        new_chains.append(chain)
    chains = new_chains
    if len(chains) == 0:
        return []
    stack = []
    output = []
    for index, chain in enumerate(chains):
        if index % 2 == 0:
            if chain[0].type not in EXPR_EXPR:
                raise_syntax_error("Expected a non-operator non-symbol token", chain[0])
            output.append(chain)
            continue
        if chain[0].type != TokenType.OPERATOR or (
                chain[0].value not in EXPR_OP
                and chain[0].value not in CMP_OP
                and chain[0].value not in CMP_COMBINE_OP):
            raise_syntax_error("Expected an operator", chain[0])
        o1 = chain
        while True:
            o2 = stack[-1] if len(stack) > 0 else None
            if o2 is None or (
                    operators[o2[0].value]["p"] <= operators[o1[0].value]["p"]
                    and (
                            operators[o2[0].value]["p"] != operators[o1[0].value]["p"]
                            or operators[o1[0].value]["a"] != "left"
                    )
            ):
                break
            output.append(stack.pop())
        stack.append(chain)

    while len(stack) > 0:
        output.append(stack.pop())

    return output


def next_token(tokens: List[Token], index: List[int]):
    index[0] += 1
    if index[0] >= len(tokens):
        return None
    return tokens[index[0]]

def peek_token(tokens: List[Token], index: List[int], offset = 1):
    i = index[0] + offset
    if i >= len(tokens):
        return None
    return tokens[i]


def parse_str(code: str):
    (tokens, macros) = tokenize(code)
    return parse(tokens, macros), macros


def parse_iterate(
        statements: List[Statement],
        tokens: List[Token],
        index: List[int],
        macros: List,
        class_names: Union[List[str], None] = None,
) -> bool:
    if class_names is None:
        class_names = []
    t0 = next_token(tokens, index)
    if t0 is None:
        return False
    if t0.type == TokenType.EOF:
        return False
    if t0.type == TokenType.EOL or t0.type == TokenType.EOE:
        return True
    if t0.value == "class":
        # class <name> { <FUNCTION STATEMENTS...> }
        t1 = next_token(tokens, index)  # name of the class
        if not t1:
            raise_syntax_error("Expected a name for the class", t0)
            return False
        if t1.value in class_names:
            raise_syntax_error("Duplicate class name", t1)
            return False
        if t1 is None or t1.type != TokenType.IDENTIFIER:
            raise_syntax_error("Expected a name for the class", t0)
            return False
        t2 = next_token(tokens, index)  # class body
        if t2 is None or not isinstance(t2, GroupToken) or t2.open.value != "{":
            raise_syntax_error("Expected a class body", t0)
            return False
        attributes: List[Tuple[str, List[Token]]] = []
        methods: List[DefineFunctionStatement] = []
        class_index = [-1]
        class_tokens = t2.children

        # A class body will include these syntax:
        # <variableType> <variableName>
        # <variableType> <variableName> = <value>
        # <functionName> { <body> }
        # <functionName>: <returnType> { <body> }
        # <functionName>( <arguments> ) { <body> }
        # <functionName>( <arguments> ): <returnType> { <body> }

        class_names.append(t1.value)

        while True:
            nt0 = next_token(class_tokens, class_index)
            if nt0 is None:
                break
            if nt0.type in ENDERS:
                continue

            nt1 = next_token(class_tokens, class_index)

            if nt0.type == TokenType.IDENTIFIER and nt1 and nt1.value == "=":
                (ind, value) = read_expression(class_tokens, class_index[0] + 1)
                if len(value) == 0:
                    raise_syntax_error(
                        "Expected an expression after '='", nt1
                    )
                    return False
                class_index[0] = ind
                attributes.append((nt0.value, value))
                continue

            class_index[0] -= 1  # for t1 being useless

            st = []
            temp_tokens = [Token("fn", TokenType.KEYWORD, 0, 2)] + class_tokens[class_index[0]:]
            temp_index = [-1]
            res = parse_iterate(st, temp_tokens, temp_index, macros, class_names)
            if not res:
                raise_syntax_error("Expected a class method declaration", nt0)
                return False
            st0 = st[0]
            if not isinstance(st0, DefineFunctionStatement):
                raise_syntax_error("Expected a class method declaration", nt0)
                return False

            methods.append(st0)
            class_index[0] += temp_index[0]

        for method in methods:
            if method.type != StatementType.DEFINE_FUNCTION:
                raise_syntax_error("Expected a class function definition", method)
                return False
        statement = DefineClassStatement(
            code=t0.code,
            start=t0.start,
            end=t2.end,
            name=t1,
            attributes=attributes,
            methods=methods,
        )
        statements.append(statement)
        return True

    if t0.value == "import":
        t1 = next_token(tokens, index)
        if t1 is None or t1.type != TokenType.STRING_LITERAL:
            raise_syntax_error("Expected a path for the import statement", t0)
            return False

        _as = None

        t2 = next_token(tokens, index)
        if t2 and t2.value == "as":
            t3 = next_token(tokens, index)
            if t3 is None or t3.type != TokenType.IDENTIFIER:
                raise_syntax_error("Expected an identifier", t2)

            _as = t3
        else:
            index[0] -= 1

        statement = ImportStatement(
            code=t0.code, start=t0.start, end=(_as or t1).end, path=t1, as_=_as
        )
        statements.append(statement)
        return True
    if (
            t0.value == "schedule"
            and index[0] + 1 < len(tokens)
            and is_time(tokens[index[0] + 1])
    ):
        # schedule 1t { <body> }
        t1 = next_token(tokens, index)
        if t1 is None or not is_time(t1):
            raise_syntax_error("Expected a time", t0)
            return False
        body = next_token(tokens, index)
        if body is None:
            raise_syntax_error("Expected an expression", t0)
            return False
        if not isinstance(body, GroupToken) or body.open.value != "{":
            raise_syntax_error("Expected an expression inside the braces", body)

        statement = ScheduleStatement(
            code=t0.code, start=t0.start, end=body.end, time=t1, body=[]
        )

        statements.append(statement)
        return True
    if t0.value == "for":
        # for <time> (<init>; <condition>; <iterator>) { <body> }
        # for (<init>; <condition>; <iterator>) { <body> }
        t1 = next_token(tokens, index)
        if t1 is None:
            raise_syntax_error("Expected an expression", t0)
            return False
        tim = None
        if is_time(t1):
            tim = t1
            t1 = next_token(tokens, index)
        if not isinstance(t1, GroupToken) or t1.open.value != "(":
            raise_syntax_error(
                "Expected an initializer, condition, and iterator inside the parentheses",
                t1 or t0,
            )
            return False
        spl = split_tokens(t1.children, ";")
        body = next_token(tokens, index)
        if not isinstance(body, GroupToken) or body.open.value != "{":
            raise_syntax_error("Expected an expression inside the parentheses", body)
            return False
        init = parse(spl[0], macros, class_names)
        if len(init) != 1:
            raise_syntax_error("Expected an expression in the initializer", t0)
        init = init[0]

        it = parse(spl[2], macros, class_names)
        if len(it) != 1:
            raise_syntax_error("Expected an expression in the iterator", t0)

        statements.append(init)

        if_condition = IfFlowStatement(
            code=t0.code,
            start=t1.start,
            end=t1.end,
            condition=spl[1],
            body=[],
            else_body=parse_str("break")[0],
        )

        loop_st_body = [if_condition]
        loop_st_body.extend(parse(body.children, macros, class_names))

        loop_st = LoopStatement(
            code=t0.code,
            start=t0.start,
            end=body.end,
            time=tim,
            body=loop_st_body,
            step=it
        )

        statements.append(loop_st)
        return True
    if t0.value == "do":
        # do { <body> } while <time> (<condition>)
        # do { <body> } until <time> (<condition>)
        # do { <body> } while(<condition>)
        # do { <body> } until(<condition>)
        body = next_token(tokens, index)
        if body is None:
            raise_syntax_error("Expected an expression", t0)
            return False
        if not isinstance(body, GroupToken) or body.open.value != "{":
            raise_syntax_error("Expected an expression inside the braces", body)
            return False
        t_while_until = next_token(tokens, index)
        if t_while_until is None or (t_while_until.value != "while" and t_while_until.value != "until"):
            raise_syntax_error("Expected 'while' or 'until'", t_while_until)
            return False
        tim = None
        condition = next_token(tokens, index)
        if condition is None:
            raise_syntax_error("Expected an expression", t_while_until)
            return False
        if is_time(condition):
            tim = condition
            condition = next_token(tokens, index)
        if not isinstance(condition, GroupToken) or condition.open.value != "(":
            raise_syntax_error(
                "Expected an expression inside the parentheses", condition
            )
            return False

        if_st = IfFlowStatement(
            code=condition.code,
            start=condition.start,
            end=condition.end,
            condition=condition.children,
            body=[],
            else_body=None,
        )

        if t_while_until.value == "while":
            if_st.body = []
            if_st.elseBody = parse_str("break")[0]
        else:
            if_st.body = parse_str("break")[0]
            if_st.elseBody = None

        loop_st = LoopStatement(
            code=t0.code,
            start=t0.start,
            end=condition.end,
            time=tim,
            body=parse(body.children, macros, class_names) + [if_st],
            step=[]
        )

        statements.append(loop_st)

        return True
    if t0.value == "while" or t0.value == "until":
        # while <time> (<condition>) { <body> }
        # until <time> (<condition>) { <body> }
        # while (<condition>) { <body> }
        # until (<condition>) { <body> }
        t1 = next_token(tokens, index)
        if t1 is None:
            raise_syntax_error("Expected an expression", t0)
            return False
        tim = None
        if is_time(t1):
            tim = t1
            t1 = next_token(tokens, index)
        if t1 is None:
            raise_syntax_error("Expected an expression", tim or t0)
            return False
        if not isinstance(t1, GroupToken) or t1.open.value != "(":
            raise_syntax_error("Expected a condition inside the parentheses", t1 or t0)
            return False
        body = next_token(tokens, index)
        if body is None:
            raise_syntax_error("Expected an expression", t1)
            return False
        if not isinstance(body, GroupToken) or body.open.value != "{":
            raise_syntax_error("Expected an expression inside the parentheses", body)
            return False

        if_st = IfFlowStatement(
            code=t0.code,
            start=t1.start,
            end=t1.end,
            condition=t1.children,
            body=[],
            else_body=None,
        )
        if t0.value == "while":
            if_st.elseBody = parse_str("break")[0]
        else:
            if_st.body = parse_str("break")[0]

        loop_st_body = [if_st]
        loop_st_body.extend(parse(body.children, macros, class_names))

        loop_st = LoopStatement(
            code=t0.code,
            start=t0.start,
            end=body.end,
            time=tim,
            body=loop_st_body,
            step=[]
        )

        statements.append(loop_st)
        return True
    if t0.value == "loop":
        # loop <time> { <body> }
        # loop { <body> }
        body = next_token(tokens, index)
        if body is None:
            raise_syntax_error("Expected an expression", t0)
            return False
        tim = None
        if is_time(body):
            tim = body
            body = next_token(tokens, index)
        if body is None:
            raise_syntax_error("Expected an expression", tim or t0)
            return False
        if not isinstance(body, GroupToken) or body.open.value != "{":
            raise_syntax_error("Expected an expression inside the parentheses", body)
            return False
        statement = LoopStatement(
            code=t0.code,
            start=t0.start,
            end=body.end,
            time=tim,
            body=parse(body.children, macros, class_names),
            step=[]
        )

        statements.append(statement)
        return True
    if (
            (t0.value == "if" or t0.value == "unless" or t0.value == "else")
            and index[0] < len(tokens) - 2
            and (t0.value == "else" or tokens[index[0] + 1].type == TokenType.GROUP)
    ):
        # if ( <condition> ) { <body> } else { <body> }
        # unless ( <condition> ) { <body> } else { <body> }
        t1 = next_token(tokens, index)  # for else this is {} or statement start
        t2 = next_token(tokens, index) if t0.value != "else" else t1
        stats = []
        if t1 is None or t2 is None:
            raise_syntax_error("Expected an expression", t0)
            return False
        if t0.value != "else" and (
                not isinstance(t1, GroupToken) or t1.open.value != "("
        ):
            raise_syntax_error("Expected an expression inside the parentheses", t1)
            return False
        if t0.value == "else":
            if len(statements) == 0 or statements[-1].type != StatementType.IF_FLOW:
                raise_syntax_error(
                    "Expected an if statement before the else statement", t0
                )
            index[0] -= 1
        if isinstance(t2, GroupToken) and t2.open.value == "{":
            stats = parse(t2.children, macros, class_names)
            end_ind = t2.end
        else:
            if t0.value != "else":
                index[0] -= 1
            parse_iterate(stats, tokens, index, macros, class_names)
            if len(stats) != 1:
                raise_syntax_error("Expected a statement", t1)
            end_ind = stats[0].end
        if t0.value == "else":
            if_st = statements[-1]
            if not isinstance(if_st, IfFlowStatement):
                raise_syntax_error(
                    "Expected an if statement before the else statement", t0
                )
                return False
            if_st.elseBody = stats
            index[0] += 1
            return True

        if not isinstance(t1, GroupToken):
            return False  # for the linter

        statement = IfFlowStatement(
            code=t0.code,
            start=t0.start,
            end=end_ind,
            condition=t1.children,
            body=stats,
            else_body=None,
        )
        statements.append(statement)
        if t0.value == "unless":
            statement.elseBody = stats
            statement.body = []
        return True

    if t0.value == "define":
        # define <type> <name>
        si = index[0]
        expr_ind = read_until_types(tokens, index[0], ENDERS)
        if expr_ind is None:
            raise_syntax_error("Unfinished define statement", t0)
        expr = tokens[si + 1:expr_ind]

        if len(expr) < 2:
            raise_syntax_error("Unfinished define statement", t0)
            return False
        if not cpl_def_from_tokens(class_names, expr[:-1]):
            raise_syntax_error("Invalid type for the define statement", t0)
            return False
        name = expr[-1]
        if (
                name.type != TokenType.IDENTIFIER
                and name.type != TokenType.SELECTOR_IDENTIFIER
        ):
            raise_syntax_error(
                "Expected a valid variable name for the define statement", name
            )
            return False
        statement = IntroduceVariableStatement(
            code=t0.code, start=t0.start, end=name.end, name=name, var_type=expr[:-1]
        )
        statements.append(statement)
        index[0] = expr_ind
        return True

    if t0.value == "fn":
        # fn <name>( <arguments> ) { <body> }
        # fn <name> { <body> }
        # fn <name>( <arguments> ): <returnType> { <body> }
        # fn <name>: <returnType> { <body> }
        t1 = next_token(tokens, index)  # function name and arguments
        t2 = next_token(tokens, index)  # function body or colon
        return_type = "auto"
        if t1 is None or (
                t1.type != TokenType.FUNCTION_CALL and t1.type != TokenType.IDENTIFIER
        ):
            raise_syntax_error(
                "Expected a function name for the function definition",
                t0,
            )
            return False
        if isinstance(t1, GroupToken) and t1.func and t1.func.value == "tick":
            raise_syntax_error(
                "Tick function does not support arguments, remove the parentheses",
                t1,
            )
            return False
        if t2 is None:
            raise_syntax_error("Expected a function body for the function", t1)
            return False
        if t2.value == ":":
            ind = read_until_curly(tokens, index[0])
            if ind is None:
                raise_syntax_error("Expected a function body for the function", t1)
                return False
            t3 = next_token(tokens, index)
            if t3 and t3.value in {"void", "auto"}:
                return_type = t3.value
                t2 = next_token(tokens, index)
            else:
                return_type = tokens[index[0]: ind]
                if not cpl_def_from_tokens(class_names, return_type):
                    raise_syntax_error(
                        "Expected a valid return type for the function", t3
                    )
                    return False
                index[0] = ind - 1
                t2 = next_token(tokens, index)

        if not isinstance(t2, GroupToken) or t2.open.value != "{":
            raise_syntax_error("Expected a function body", t2)
            return False

        statement = DefineFunctionStatement(
            code=t1.code,
            start=t0.start,
            end=t2.end,
            name=t1.func if isinstance(t1, GroupToken) and t1.func else t1,
            body=parse(t2.children, macros, class_names),
            arguments=[],
            returns=return_type,
        )
        if isinstance(t1, GroupToken) and t1.func:
            spl = split_tokens(t1.children)
            for s in spl:
                o_chain = chain_tokens(s)
                chain = o_chain
                if len(chain) == 3 and chain[0][0].value == "$":
                    chain = chain[1:]
                if len(chain) != 2:
                    raise_syntax_error("Expected a valid function argument", s[0])
                c_type = chain[0]
                type_def = cpl_def_from_tokens(class_names, c_type)
                if not type_def:
                    raise_syntax_error(
                        "Expected a valid type for the argument", c_type[0]
                    )
                statement.arguments.append(o_chain)

        statements.append(statement)
        return True

    if t0.type == TokenType.KEYWORD and t0.value == "return":
        # return <expression>
        expr_tokens = []
        ind = index[0] + 1
        if ind <= len(tokens) and tokens[ind].type not in ENDERS:
            (ind, expr_tokens) = read_expression(tokens, ind)
            if len(expr_tokens) == 0:
                raise_syntax_error(
                    "Expected an expression after the return keyword", t0
                )
        statement = ReturnStatement(
            code=t0.code,
            start=t0.start,
            end=t0.end if len(expr_tokens) == 0 else expr_tokens[-1].end,
            keyword=t0,
            expr_tokens=expr_tokens,
        )
        statements.append(statement)
        index[0] = ind
        return True

    if t0.type == TokenType.KEYWORD and (t0.value == "break" or t0.value == "continue"):
        # break
        # continue
        statement = (BreakStatement if t0.value == "break" else ContinueStatement)(
            code=t0.code, start=t0.start, end=t0.end, keyword=t0
        )
        statements.append(statement)
        return True

    if t0.value in EXECUTE_MACROS:
        i_start = index[0]
        is_exec_macro = False
        for tk in tokens[i_start:]:
            if isinstance(tk, GroupToken) and tk.open.value == "{":
                is_exec_macro = True
                break
        if is_exec_macro:
            # as @a at @s ... { <body> }
            got_tokens = [t0]
            body_g: GroupToken
            while True:
                t = next_token(tokens, index)
                if t is None:
                    raise_syntax_error("Unexpected end of file", got_tokens[-1])
                    assert False
                if isinstance(t, GroupToken) and t.open.value == "{":
                    body_g = t
                    break
                got_tokens.append(t)

            statement = ExecuteMacroStatement(
                code=t0.code,
                start=t0.start,
                end=got_tokens[-1].end,
                command=t0.code[got_tokens[0].start: got_tokens[-1].end],
                body=parse(body_g.children, macros, class_names),
            )
            statements.append(statement)
            return True

    (ind, expr_tokens) = read_expression(tokens, index[0])
    if ind is None:
        raise_syntax_error("Unexpected token", t0)
    statement = InlineStatement(
        code=t0.code, start=t0.start, end=expr_tokens[-1].end, expr_tokens=expr_tokens
    )
    statements.append(statement)
    index[0] = ind
    return True


def parse(
        tokens: List[Token], macros: List, class_names: List[str] = None
):
    if class_names is None:
        class_names = []
    statements: List[Statement] = []
    index = [-1]

    while parse_iterate(statements, tokens, index, macros, class_names):
        pass
    return statements



from typing import Union, List

from ._base import CompileTimeValue
from ..error import raise_syntax_error
from ..tokenizer import Token
from ..utils import INT_TYPE, get_uuid, FLOAT_PREC, FLOAT_TYPE, inv_cmp


class CplScore(CompileTimeValue):
    def __init__(
            self,
            token: Union[Token, None],
            location: str,
            type: str | None = None,
    ):
        super().__init__(token)
        if type is None:
            type = location.split("_")[0]  # type: ignore
        self.location = location
        if type == "int":
            self.unique_type = INT_TYPE
        elif type == "float":
            self.unique_type = FLOAT_TYPE

    def get_data_str(self, ctx):
        return "from " + self.cache(ctx, force="nbt").location

    def as_int(self, ctx, score_loc=None):
        if self.unique_type.type == "int":
            return self
        eid = score_loc or f"int_{get_uuid()} __temp__"
        ctx.file.append(f"scoreboard players operation {eid} = {self.location}")
        ctx.file.append(f"scoreboard players operation {eid} /= FLOAT_PREC __temp__")
        return CplScore(self.token, eid)

    def as_float(self, ctx, score_loc=None):
        if self.unique_type.type == "float":
            return self
        eid = score_loc or f"float_{get_uuid()} __temp__"
        ctx.file.append(f"scoreboard players operation {eid} = {self.location}")
        ctx.file.append(f"scoreboard players operation {eid} *= FLOAT_PREC __temp__")
        return CplScore(self.token, eid)

    def _force_float(self, ctx):  # ONLY USE THIS ON RECENTLY CACHED SCORES!
        if self.unique_type == FLOAT_TYPE:
            return
        self.unique_type = FLOAT_TYPE
        ctx.file.append(f"scoreboard players operation {self.location} *= FLOAT_PREC __temp__")

    def _force_int(self, ctx):  # ONLY USE THIS ON RECENTLY CACHED SCORES!
        if self.unique_type == INT_TYPE:
            return
        self.unique_type = INT_TYPE
        ctx.file.append(f"scoreboard players operation {self.location} /= FLOAT_PREC __temp__")

    def _force_type(self, ctx, type):  # ONLY USE THIS ON RECENTLY CACHED SCORES!
        if type == "int":
            return self._force_int(ctx)
        self._force_float(ctx)

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        if force == "nbt":
            t_have = self.unique_type.type
            mul = "int 1"
            if force_t == "float" and t_have == "int":
                mul = f"int {FLOAT_PREC}"
            if t_have == "float":
                mul = f"float {1 / FLOAT_PREC:.7f}"
            ctx.file.append(f"execute store result {nbt_loc} {mul} run scoreboard players get {self.location}")
            return val_nbt(self.token, nbt_loc, force_t or self.unique_type)
        if not score_loc:
            score_loc = f"{self.unique_type.type}_{get_uuid()} __temp__"
        ctx.file.append(
            f"scoreboard players operation {score_loc} = {self.location}"
        )
        score = CplScore(self.token, score_loc, self.unique_type.type)
        if force_t:
            score._force_type(ctx, force_t)
        return score

    def _call_index(self, ctx, index: str, arguments: List[CompileTimeValue], token):
        t = self.unique_type.type
        if index == "sqrt":
            return ctx.transpiler.builtin_vars["Math"].value._call_index(ctx, "isqrt" if t == "int" else "sqrt", [self],
                                                                         token)
        if index == "cbrt":
            return ctx.transpiler.builtin_vars["Math"].value._call_index(ctx, "cbrt", [self], token)
        if index == "int":
            if len(arguments) > 0:
                raise_syntax_error("Expected 0 arguments for <score>.int()", self.token)
                assert False
            return self.as_int(ctx)
        if index == "float":
            if len(arguments) > 0:
                raise_syntax_error("Expected 0 arguments for <score>.int()", self.token)
                assert False
            return self.as_float(ctx)
        return None

    def __score_help(self, cpl):
        t_want = self.unique_type.type
        t_have = cpl.unique_type.type
        v = cpl.value
        if t_want == "int" and t_have == "float":
            v = int(v)
        if t_want == "float":
            v = int(v * FLOAT_PREC)
        return v

    def _set(self, ctx, cpl):
        if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
            v = self.__score_help(cpl)
            ctx.file.append(f"scoreboard players set {self.location} {v}")
            return self
        t_want = self.unique_type.type
        t_have = cpl.unique_type.type
        if not isinstance(cpl, CplScore) and not isinstance(cpl, CplIntNBT) and not isinstance(cpl, CplFloatNBT):
            return None
        if isinstance(cpl, CplScore):
            ctx.file.append(f"scoreboard players operation {self.location} = {cpl.location}")
        else:
            if t_want == "int":
                ctx.file.append(f"execute store result score {self.location} run data get {cpl.location}")
            else:
                ctx.file.append(f"execute store result score {self.location} run data get {cpl.location} {FLOAT_PREC}")
            return self
        if t_want == "int" and t_have == "float":
            ctx.file.append(f"scoreboard players operation {self.location} /= FLOAT_PREC __temp__")
        if t_want == "float" and t_have == "int":
            ctx.file.append(f"scoreboard players operation {self.location} *= FLOAT_PREC __temp__")
        return self

    def _set_add(self, ctx, cpl):
        if cpl.is_lit_eq(0):
            return self
        if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
            ctx.file.append(f"scoreboard players add {self.location} {self.__score_help(cpl)}")
            return self
        if isinstance(cpl, CplScore):
            t_want = self.unique_type.type
            t_have = cpl.unique_type.type
            if t_want != t_have:
                cpl = cpl.cache(ctx, force=t_want)
            ctx.file.append(f"scoreboard players operation {self.location} += {cpl.location}")
            return self
        if not isinstance(cpl, CplIntNBT) and not isinstance(cpl, CplFloatNBT):
            return None
        return self._set_add(ctx, cpl.cache(ctx, force="score", force_t=self.unique_type.type))

    def _set_sub(self, ctx, cpl):
        if cpl.is_lit_eq(0):
            return self
        if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
            ctx.file.append(f"scoreboard players remove {self.location} {self.__score_help(cpl)}")
            return self
        if isinstance(cpl, CplScore):
            t_want = self.unique_type.type
            t_have = cpl.unique_type.type
            if t_want != t_have:
                cpl = cpl.cache(ctx, force_t=t_want)
            ctx.file.append(f"scoreboard players operation {self.location} -= {cpl.location}")
            return self
        if not isinstance(cpl, CplIntNBT) and not isinstance(cpl, CplFloatNBT):
            return None
        return self._set_sub(ctx, cpl.cache(ctx, force="score", force_t=self.unique_type.type))

    def _set_mul(self, ctx, cpl):
        if cpl.is_lit_eq(1):
            return self
        if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
            return self._set_mul(ctx, cpl.cache(ctx, force="score", force_t=self.unique_type.type))
        if isinstance(cpl, CplScore):
            ctx.file.append(f"scoreboard players operation {self.location} *= {cpl.location}")
            if cpl.unique_type.type == "float":
                ctx.file.append(f"scoreboard players operation {self.location} /= FLOAT_PREC __temp__")
            return self
        if not isinstance(cpl, CplIntNBT) and not isinstance(cpl, CplFloatNBT):
            return None
        return self._set_mul(ctx, cpl.cache(ctx, force="score", force_t=self.unique_type.type))

    def _set_div(self, ctx, cpl):
        if cpl.is_lit_eq(0):
            raise_syntax_error("Divide by zero", self.token)
        if cpl.is_lit_eq(1):
            return self
        if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
            return self._set_div(ctx, cpl.cache(ctx, force="score", force_t=self.unique_type.type))
        if isinstance(cpl, CplScore):
            if cpl.unique_type.type == "float":
                ctx.file.append(f"scoreboard players operation {self.location} *= FLOAT_PREC __temp__")
            ctx.file.append(f"scoreboard players operation {self.location} /= {cpl.location}")
            return self
        if not isinstance(cpl, CplIntNBT) and not isinstance(cpl, CplFloatNBT):
            return None
        return self._set_div(ctx, cpl.cache(ctx, force="score", force_t=self.unique_type.type))

    def _set_mod(self, ctx, cpl):
        if cpl.is_lit_eq(0):
            raise_syntax_error("Modulo by zero", self.token)
        if cpl.is_lit_eq(1) and self.unique_type.type == "int":
            return self
        if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
            return self._set_mod(ctx, cpl.cache(ctx, force="score", force_t=self.unique_type.type))
        if isinstance(cpl, CplScore):
            t_want = self.unique_type.type
            t_have = cpl.unique_type.type
            if t_want != t_have:
                cpl = cpl.cache(ctx, force_t=t_want)
            ctx.file.append(f"scoreboard players operation {self.location} %= {cpl.location}")
            return self
        if not isinstance(cpl, CplIntNBT) and not isinstance(cpl, CplFloatNBT):
            return None
        return self._set_mod(ctx, cpl.cache(ctx, force="score", force_t=self.unique_type.type))

    def _set_pow(self, ctx, cpl):
        if cpl.unique_type.type == "float":
            return None
        return self._set(ctx,
                         ctx.transpiler.builtin_vars["Math"].value._call_index(ctx, "ipow", [self, cpl], self.token))

    def _add(self, ctx, cpl):
        if cpl.is_lit_eq(0):
            return self
        c = self._cache(ctx)
        if self.unique_type.type == "int" and cpl.unique_type.type == "float":
            c._force_float(ctx)
        return c._set_add(ctx, cpl)

    def _sub(self, ctx, cpl):
        if cpl.is_lit_eq(0):
            return self
        c = self._cache(ctx)
        if self.unique_type.type == "int" and cpl.unique_type.type == "float":
            c._force_float(ctx)
        return c._set_sub(ctx, cpl)

    def _mul(self, ctx, cpl):
        if cpl.is_lit_eq(0):
            return cpl
        if cpl.is_lit_eq(1):
            return self
        c = self._cache(ctx)
        if self.unique_type.type == "int" and cpl.unique_type.type == "float":
            c._force_float(ctx)
        return c._set_mul(ctx, cpl)

    def _div(self, ctx, cpl):
        if cpl.is_lit_eq(0):
            raise_syntax_error("Divide by zero", self.token)
        if cpl.is_lit_eq(1):
            return cpl
        c = self._cache(ctx)
        if self.unique_type.type == "int" and cpl.unique_type.type == "float":
            c._force_float(ctx)
        return c._set_div(ctx, cpl)

    def _mod(self, ctx, cpl):
        if cpl.is_lit_eq(0):
            raise_syntax_error("Modulo by zero", self.token)
        if cpl.is_lit_eq(1) and self.unique_type.type == "int":
            return cpl
        c = self._cache(ctx)
        if self.unique_type.type == "int" and cpl.unique_type.type == "float":
            c._force_float(ctx)
        return c._set_mod(ctx, cpl)

    def _pow(self, ctx, cpl):
        if cpl.unique_type.type == "float":
            return None
        return ctx.transpiler.builtin_vars["Math"].value._call_index(ctx, "ipow", [self, cpl], self.token)

    def _eq_neq(self, ctx, cpl, is_eq):
        return self._cmp(ctx, cpl, "==" if is_eq else "!=")

    def _cmp(self, ctx, cpl, op):
        if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
            if op == "==" or op == "!=":
                return cpl._eq_neq(ctx, self, op == "==")
            return cpl._cmp(ctx, self, inv_cmp(op))

        if isinstance(cpl, CplIntNBT) or isinstance(cpl, CplFloatNBT):
            cpl = cpl.cache(ctx, force="score", force_t=self.unique_type.type)

        if isinstance(cpl, CplScore):
            op_if = "unless" if op == "!=" else "if"
            eid = f"int_{get_uuid()} __temp__"
            t_want = self.unique_type.type
            t_have = cpl.unique_type.type
            if t_want != t_have:
                cpl = cpl.cache(ctx, force_t=t_want)
            if op == "==" or op == "!=":
                op = "="
            ctx.file.append(f"scoreboard players set {eid} 0")
            ctx.file.append(f"execute {op_if} score {self.location} {op} {cpl.location} "
                            f"run scoreboard players set {eid} 1")
            return CplScore(self.token, eid)
        return None

    def _and(self, ctx, cpl):
        if isinstance(cpl, CplSelector):
            return cpl._and(ctx, self)
        eid = f"int_{get_uuid()} __temp__"
        ctx.file.append(f"scoreboard players set {eid} 0")
        ctx.file.append(f"execute "
                        f"unless score {self.location} matches 0..0 "
                        f"unless score {cpl.location} matches 0..0 "
                        f"run scoreboard players set {eid} 1")
        return CplScore(self.token, eid)

    def _or(self, ctx, cpl):
        if isinstance(cpl, CplSelector):
            return cpl._and(ctx, self)
        eid = f"int_{get_uuid()} __temp__"
        ctx.file.append(f"scoreboard players set {eid} 0")
        ctx.file.append(f"execute "
                        f"unless score {self.location} matches 0..0 "
                        f"run scoreboard players set {eid} 1")
        ctx.file.append(f"execute "
                        f"unless score {cpl.location} matches 0..0 "
                        f"run scoreboard players set {eid} 1")
        return CplScore(self.token, eid)

    def tellraw_object(self, ctx):
        if self.unique_type.type == "int":
            ls = self.location.split(" ")
            return {
                "score": {
                    "name": ls[0],
                    "objective": ls[1]
                }
            }
        eid = get_uuid()
        ctx.file.append(
            f"execute store result storage temp _{eid} float {1 / FLOAT_PREC:.7f} run scoreboard players get {self.location}")
        return {
            "storage": "temp",
            "nbt": f"_{eid}"
        }


from .float import CplFloat
from .int import CplInt
from .nbt import val_nbt
from .nbtfloat import CplFloatNBT
from .nbtint import CplIntNBT
from .selector import CplSelector



import json
from typing import List

from ..cpl._base import CompileTimeValue
from ..cpl.score import CplScore
from ..cpl.string import CplString
from ..transpiler import FunctionDeclaration, TranspilerContext, add_lib
from ..utils import get_uuid


def get_fn_or_none(ctx, arr, index, token):
    if len(arr) <= index:
        return None
    if isinstance(arr[index], CplString):
        return ctx.transpiler.get_fn(arr[index].value, [], token).file_name
    return None


def lib_raycast(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    ctx.transpiler.dp_files[f"tags/block{ctx.transpiler.s}/__raycast__/default_raycast_pass.json"] = json.dumps({
        "values": [
            "minecraft:short_grass" if ctx.transpiler.pack_format >= 48 else "minecraft:grass",
            "minecraft:air",
            "minecraft:void_air",
            "minecraft:cave_air",
            "minecraft:water",
            "minecraft:lava",
            "#minecraft:small_flowers",
            "#minecraft:tall_flowers",
            "#minecraft:small_dripleaf_placeable",
            "minecraft:fern",
            "minecraft:fire",
            "minecraft:tall_grass",
            "minecraft:large_fern",
            "minecraft:vine",
            "minecraft:twisting_vines",
            "minecraft:twisting_vines_plant",
            "minecraft:weeping_vines",
            "minecraft:weeping_vines_plant",
            "#minecraft:crops",
            "#minecraft:saplings",
            "#minecraft:signs",
            "minecraft:attached_melon_stem",
            "minecraft:attached_pumpkin_stem",
            "minecraft:nether_wart",
            "minecraft:sweet_berry_bush",
            "minecraft:cocoa",
            "minecraft:sugar_cane",
            "minecraft:seagrass",
            "minecraft:tall_seagrass",
            "minecraft:redstone_wire",
            "minecraft:rail",
            "minecraft:powered_rail",
            "minecraft:activator_rail",
            "minecraft:detector_rail",
            "minecraft:torch",
            "minecraft:soul_torch",
            "minecraft:redstone_torch",
            "minecraft:glow_lichen"
        ]
    })
    raycast_id = get_uuid()
    iter_eid = f"int_{raycast_id} __temp__"
    success_eid = f"int_raycast_success_{raycast_id} __temp__"
    stop_iter = f"scoreboard players set {iter_eid} -1"
    ns = ctx.transpiler.pack_namespace

    entity_fn = get_fn_or_none(ctx, args, 0, token)
    block_fn = get_fn_or_none(ctx, args, 1, token)
    step_fn = get_fn_or_none(ctx, args, 2, token)
    fail_fn = get_fn_or_none(ctx, args, 3, token)

    entity_file = f"__raycast__/{raycast_id}/collide_entity" if entity_fn else ""
    block_file = f"__raycast__/{raycast_id}/collide_block" if block_fn else ""
    fail_file = f"__raycast__/{raycast_id}/fail" if fail_fn else ""

    ctx.file.extend([
        f"scoreboard players set {success_eid} 1",
        "tag @s add __raycast__self__",
        f"scoreboard players set {iter_eid} 1000",
        f"execute anchored eyes positioned ^ ^ ^0.1 run function {ns}:__raycast__/{raycast_id}/loop",
        "tag @s remove __raycast__self__"
    ])
    ctx.transpiler.files[f"__raycast__/{raycast_id}/loop"] = [
        f"function {ns}:{step_fn}" if step_fn else "",

        f"execute "
        f"positioned ~-0.95 ~-0.95 ~-0.95 "
        f"as @e[dx=0,tag=!__raycast__self__] "
        f"positioned ~0.9 ~0.9 ~0.9 "
        f"if entity @s[dx=0] "
        f"positioned ~0.05 ~0.05 ~0.05 "
        f"run function {ns}:{entity_file}" if entity_fn else "",

        f"execute "
        f"if score {iter_eid} matches 1.. "
        f"run scoreboard players remove {iter_eid} 1",

        f"execute "
        f"unless block ~ ~ ~ #{ns}:__raycast__/default_raycast_pass "
        f"run function {ns}:{block_file}" if block_fn else "",

        f"execute "
        f"unless score {iter_eid} matches 1.. "
        f"run function {ns}:{fail_file}" if fail_fn else "",

        f"execute "
        f"if score {iter_eid} matches 1.. "
        f"positioned ^ ^ ^0.1 "
        f"run function {ns}:__raycast__/{raycast_id}/loop"
    ]
    if entity_fn:
        ctx.transpiler.files[f"__raycast__/{raycast_id}/collide_entity"] = [
            stop_iter,
            f"function {ns}:{entity_fn}"
        ]
    if block_fn:
        ctx.transpiler.files[f"__raycast__/{raycast_id}/collide_block"] = [
            stop_iter,
            f"function {ns}:{block_fn}"
        ]
    if fail_fn:
        ctx.transpiler.files[f"__raycast__/{raycast_id}/fail"] = [
            f"scoreboard players set {success_eid} 0",
            f"function {ns}:{fail_fn}"
        ]
    return CplScore(token, success_eid)


add_lib(FunctionDeclaration(
    type="python-cpl",
    name="raycast",
    function=lib_raycast
))



from typing import List

from ..cpl.string import CplString
from ..utils import VariableDeclaration
from ..cpl._base import CompileTimeValue
from ..cpl.int import CplInt
from ..cpl.selector import CplSelector
from ..error import raise_syntax_error
from ..transpiler import FunctionDeclaration, TranspilerContext, add_lib, CustomCplObject


def _lib_print(ctx: TranspilerContext, args: List[CompileTimeValue], token, prefix: str, name: str):
    if len(args) < 1:
        raise_syntax_error(f"Invalid arguments. Expected usage: {name}(selector, ...anything)", token)
    if not isinstance(args[0], CplSelector):
        args.insert(0, CplSelector(token, "@a"))
    prefix = prefix.replace("@", args[0].value)
    if len(args) == 2:
        ctx.file.append(f"{prefix} {args[1].tellraw_object_str(ctx)}")
    else:
        ctx.file.append(f"{prefix} {", ".join(x.tellraw_object_str(ctx) for x in args[1:])}")
    return CplInt(token, 0)


def lib_print(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return _lib_print(ctx, args, token, "tellraw @", "print")


def lib_print_title(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return _lib_print(ctx, args, token, "title @ title", "printTitle")


def lib_print_subtitle(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return _lib_print(ctx, args, token, "title @ subtitle", "printSubtitle")


def lib_print_actionbar(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return _lib_print(ctx, args, token, "title @ actionbar", "printActionbar")


class CplFormat(CompileTimeValue):
    def __init__(self, token, fmt, value: CompileTimeValue):
        super().__init__(token)
        self.fmt = fmt
        self.value = value

    def _raw_call_index(self, ctx, index, args, token):
        if index == "hover":
            if len(args) != 2:
                raise_syntax_error("Expected 2 arguments for fmt.hover()", token)
            if "hoverEvent" in self.fmt:
                raise_syntax_error("fmt.hover() already set", token)
            args[0] = args[0].cpl(ctx)
            self.fmt["hoverEvent"] = {"action": args[0].value, "contents": lambda: args[1].value}
            return self
        return None

    def _call_index(self, ctx, index, args, token):
        if index == "insertion":
            if len(args) != 1:
                raise_syntax_error("Expected 1 argument for fmt.insertion()", token)
            if "insertion" in self.fmt:
                raise_syntax_error("fmt.insertion() already set", token)
            if not isinstance(args[0], CplString):
                raise_syntax_error("Expected a string argument for fmt.insertion()", token)
            self.fmt["insertion"] = args[0].value
            return self
        if index == "extend":
            if len(args) != 1:
                raise_syntax_error("Expected 1 argument for fmt.extend()", token)
            if not isinstance(args[0], CplFormat):
                raise_syntax_error("Expected a RadonFormat argument for fmt.extend()", token)
            if "extra" not in self.fmt:
                self.fmt["extra"] = []
            self.fmt["extra"].append(args[0].tellraw_object(ctx))
            return self
        if index == "click":
            if len(args) != 2:
                raise_syntax_error("Expected 2 arguments for fmt.click()", token)
            if "clickEvent" in self.fmt:
                raise_syntax_error("fmt.click() already set", token)
            if not isinstance(args[0], CplString) or not isinstance(args[1], CplString):
                raise_syntax_error("Expected 2 string arguments for fmt.click()", token)
            self.fmt["clickEvent"] = {"action": args[0].value, "value": args[1].value}
            return self
        return None

    def tellraw_object(self, ctx):
        obj = {}
        obj.update(self.value.tellraw_object(ctx))
        obj.update(self.fmt)
        return obj


def lib_fmt(ctx, args, token):
    fmt = {}
    for arg in args[:-1]:
        value = arg.value
        if (value in {"black", "dark_blue", "dark_green", "dark_aqua", "dark_red", "dark_purple", "gold", "gray",
                      "dark_gray", "blue", "green", "aqua", "red", "light_purple", "yellow", "white"}
                or value[0] == "#"):
            if "color" in fmt:
                raise_syntax_error("Cannot use multiple colors", token)
            fmt["color"] = value
        if value in {"bold", "italic", "underlined", "strikethrough", "obfuscated"}:
            fmt[value] = True

    val = args[-1].cpl(ctx)
    while isinstance(val, CplFormat):
        fmt.update(val.fmt)
        val = val.value
    return CplFormat(token, fmt, val)


add_lib(FunctionDeclaration(
    type="python-cpl",
    name="print",
    arguments=[],
    function=lib_print
), FunctionDeclaration(
    type="python-cpl",
    name="printTitle",
    arguments=[],
    function=lib_print_title
), FunctionDeclaration(
    type="python-cpl",
    name="printSubtitle",
    arguments=[],
    function=lib_print_subtitle
), FunctionDeclaration(
    type="python-cpl",
    name="printActionbar",
    arguments=[],
    function=lib_print_actionbar
), VariableDeclaration(
    name="fmt",
    type=CustomCplObject(),
    constant=True
), FunctionDeclaration(
    type="python-raw",
    name="fmt",
    function=lib_fmt
))
